import numpy as np
from scipy.optimize import curve_fit
from prolint2.metrics.base import FittingFunction


class BiExpoFittingFunction(FittingFunction):
    name = "bi_expo"
    p0 = [1, 1.0, 1.0, 1.0]
    maxfev = 1000000

    def compute(self, x, k1, k2, A, B):
        exp1 = np.exp(np.clip(-k1 * x, None, 700))
        exp2 = np.exp(np.clip(-k2 * x, None, 700))
        return A * exp1 + B * exp2

    def get_koff(self, popt):
        ks = [abs(k) for k in popt[:2]]
        return np.min(ks)


class MonoExpoFittingFunction(FittingFunction):
    name = "mono_expo"
    p0 = [1, 1]
    maxfev = 1000000

    def compute(self, x, k, A):
        exp = np.exp(np.clip(-k * x, None, 700))
        return A * exp

    def get_koff(self, popt):
        return abs(popt[0])


class PolynomialFittingFunction(FittingFunction):
    name = "poly"
    p0 = [1, 1, 1, 1]
    maxfev = 1000000

    def __init__(self, degree=None):
        self.degree = degree if degree is not None else 1

    def compute(self, x, *params):
        return np.polyval(params, x)

    def get_koff(self, popt):
        # Calculate koff based on the polynomial coefficients
        raise NotImplementedError(
            "koff calculation for polynomial fitting function is not implemented yet"
        )

    def fit(self, x_data, y_data, **kwargs):
        degree = kwargs.pop("degree", None)
        if degree is not None:
            self.degree = degree
        if "p0" not in kwargs and self.degree is not None:
            kwargs["p0"] = [1] * (self.degree + 1)
        kwargs.pop("degree", None)
        popt, _ = curve_fit(self.compute, x_data, y_data, **kwargs)
        return popt


class FittingFunctionFactory:
    @staticmethod
    def get_fitting_function(name):
        try:
            return FittingFunction.registry[name]()
        except KeyError:
            raise ValueError(f"Invalid fitting function name: {name}")



from abc import ABC, abstractmethod
from typing import Union

from prolint2.metrics.formatters import OutputFormat, SingleOutputFormat
from prolint2.metrics.registries import MetricRegistry


class BaseOutputFormatConverter(ABC):
    def __init__(
        self,
        output_format: OutputFormat,
        metric_type: Union[str, int],
        metric_registry: MetricRegistry,
    ):
        self.output_format = output_format
        self.metric_type = metric_type
        self.metric_registry = metric_registry

    @abstractmethod
    def convert(self) -> OutputFormat:
        pass


class DefaultToSingleConverter(BaseOutputFormatConverter):
    def convert(self) -> SingleOutputFormat:
        single_output_format = SingleOutputFormat()
        metric_class_name = self.metric_registry.get_metric(self.metric_type).__name__

        if not isinstance(self.metric_type, str):
            raise ValueError(
                "The metric type must be an string when using the DefaultToSingleConverter"
            )

        for residue_id, lipid_data in self.output_format.items():
            for lipid_id, metric_data in lipid_data.items():
                single_output_format.store_result(
                    residue_id,
                    lipid_id,
                    self.metric_type,
                    metric_data[metric_class_name],
                )

        return single_output_format


class CustomToSingleConverter(BaseOutputFormatConverter):
    def convert(self) -> SingleOutputFormat:
        single_output_format = SingleOutputFormat()

        if not isinstance(self.metric_type, int):
            raise ValueError(
                "The metric type must be an integer when using the CustomToSingleConverter"
            )

        for residue_id, lipid_data in self.output_format.items():
            for lipid_id, metric_values in lipid_data.items():
                single_output_format.store_result(
                    residue_id,
                    lipid_id,
                    self.metric_type,
                    metric_values[self.metric_type],
                )

        return single_output_format



r"""UFCC sample data
======================================================
:Authors: Daniel P. Ramirez & Besian I. Sejdiu
:Year: 2022
:Copyright: MIT License
"""

import os
import requests
import tqdm


def file_download(url: str, filename: str, desc: str):
    with open(filename, "wb") as f:
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            total = int(r.headers.get("content-length", 0))

            tqdm_params = {
                "desc": url,
                "total": total,
                "miniters": 1,
                "unit": "B",
                "unit_scale": True,
                "unit_divisor": 1024,
                "desc": desc,
            }
            with tqdm.tqdm(**tqdm_params) as pb:
                for chunk in r.iter_content(chunk_size=8192):
                    pb.update(len(chunk))
                    f.write(chunk)


class GIRKDataSample:
    """
    Class to add sample data in order to test the installation of the package

    It uses the coordinates and trajectory of a simulation of the GIRK channel using the MARTINI force field.

    Example
    -------
    Import the prolint2 library and use the sample data as follows::

        from prolint2 import PL2
        from prolint2.sampledata import GIRK

        target_system = PL2(GIRK.coordinates, GIRK.trajectory)

    """

    def __init__(self):
        self.path = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), "data/GIRK/"
        )
        if (
            os.path.isfile(os.path.join(self.path, "coordinates.gro")) is False
            or os.path.isfile(os.path.join(self.path, "trajectory.xtc")) is False
        ):
            file_download(
                "https://prolint.github.io/ProLintData/GIRK/mid/coordinates.gro",
                os.path.join(self.path, "coordinates.gro"),
                "Downloading GIRK coordinates...",
            )
            file_download(
                "https://prolint.github.io/ProLintData/GIRK/mid/trajectory.xtc",
                os.path.join(self.path, "trajectory.xtc"),
                "Downloading GIRK trajectory...",
            )
            self.coordinates = os.path.join(self.path, "coordinates.gro")
            self.trajectory = os.path.join(self.path, "trajectory.xtc")
        else:
            self.coordinates = os.path.join(self.path, "coordinates.gro")
            self.trajectory = os.path.join(self.path, "trajectory.xtc")
        self.contacts = os.path.join(self.path, "contacts.csv")


class COX1DataSample:
    """
    Class to add sample data in order to test the installation of the package

    It uses the coordinates and trajectory of a simulation of the COX1 enzyme using the MARTINI force field.

    Example
    -------
    Import the prolint2 library and use the sample data as follows::

        from prolint2 import PL2
        from prolint2.sampledata import COX1

        target_system = PL2(COX1.coordinates, COX1.trajectory)

    """

    def __init__(self):
        self.path = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), "data/COX1/"
        )
        if (
            os.path.isfile(os.path.join(self.path, "coordinates.gro")) is False
            or os.path.isfile(os.path.join(self.path, "trajectory.xtc")) is False
        ):
            file_download(
                "https://prolint.github.io/ProLintData/COX1/mid/coordinates.gro",
                os.path.join(self.path, "coordinates.gro"),
                "Downloading COX1 coordinates...",
            )
            file_download(
                "https://prolint.github.io/ProLintData/COX1/mid/trajectory.xtc",
                os.path.join(self.path, "trajectory.xtc"),
                "Downloading COX1 trajectory...",
            )
            self.coordinates = os.path.join(self.path, "coordinates.gro")
            self.trajectory = os.path.join(self.path, "trajectory.xtc")
        else:
            self.coordinates = os.path.join(self.path, "coordinates.gro")
            self.trajectory = os.path.join(self.path, "trajectory.xtc")
        self.contacts = os.path.join(self.path, "contacts.csv")


class SMODataSample:
    """
    Class to add sample data in order to test the installation of the package

    It uses the coordinates and trajectory of a simulation of the SMO receptor using the MARTINI force field.

    Example
    -------
    Import the prolint2 library and use the sample data as follows::

        from prolint2 import PL2
        from prolint2.sampledata import SMO

        target_system = PL2(SMO.coordinates, SMO.trajectory)

    """

    def __init__(self):
        self.path = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), "data/SMO/"
        )
        if (
            os.path.isfile(os.path.join(self.path, "coordinates.gro")) is False
            or os.path.isfile(os.path.join(self.path, "trajectory.xtc")) is False
        ):
            file_download(
                "https://prolint.github.io/ProLintData/Smoothened/mid/coordinates.gro",
                os.path.join(self.path, "coordinates.gro"),
                "Downloading SMO coordinates...",
            )
            file_download(
                "https://prolint.github.io/ProLintData/Smoothened/mid/trajectory.xtc",
                os.path.join(self.path, "trajectory.xtc"),
                "Downloading SMO trajectory...",
            )
            self.coordinates = os.path.join(self.path, "coordinates.gro")
            self.trajectory = os.path.join(self.path, "trajectory.xtc")
        else:
            self.coordinates = os.path.join(self.path, "coordinates.gro")
            self.trajectory = os.path.join(self.path, "trajectory.xtc")
        self.contacts = os.path.join(self.path, "contacts.csv")



import warnings
import pandas as pd
import pytest

import MDAnalysis as mda
from typing import Literal, get_args

from prolint2 import Universe
from prolint2.core.groups import ExtendedAtomGroup
from prolint2.sampledata import GIRKDataSample, COX1DataSample, SMODataSample

GIRK = GIRKDataSample()
COX1 = COX1DataSample()
SMO = SMODataSample()

warnings.filterwarnings("ignore")

TimeUnitLiteral = Literal["fs", "ps", "ns", "us", "ms", "s"]

# Build VALID_UNITS from TimeUnitLiteral
VALID_UNITS = get_args(TimeUnitLiteral)


@pytest.fixture(scope="module")
def universes():
    return [
        Universe(GIRK.coordinates, GIRK.trajectory),
        Universe(COX1.coordinates, COX1.trajectory),
        Universe(SMO.coordinates, SMO.trajectory),
    ]


class TestUniverse:
    def test_init_with_universe(self, universes):
        for u in universes:
            assert isinstance(u, Universe)
            assert isinstance(u, mda.Universe)

    def test_query_property(self, universes):
        for u in universes:
            assert u.query is not None
            assert isinstance(u.query, ExtendedAtomGroup)

    def test_database_property(self, universes):
        for u in universes:
            assert u.database is not None
            assert isinstance(u.database, ExtendedAtomGroup)

    # def test_compute_contacts(self, universes):
    #     dict_of_contacts = {0: GIRK, 1: COX1, 2: SMO}
    #     for i, u in enumerate(universes):
    #         contacts = u.compute_contacts(cutoff=7)
    #         calc_df = contacts.create_dataframe(u.trajectory.n_frames)
    #         ref_df = pd.read_csv(dict_of_contacts[i].contacts, index_col=[0, 1])
    #         ref_df.columns = ref_df.columns.astype(int)
    #         ref_df = ref_df.astype("int8")
    #         pd.testing.assert_frame_equal(calc_df, ref_df)

    # def test_load_contacts_from_file(self, universes):
    #     dict_of_contacts = {0: GIRK, 1: COX1, 2: SMO}
    #     for i, u in enumerate(universes):
    #         contacts = u.load_contacts_from_file(dict_of_contacts[i].contacts)
    #         calc_df = contacts.create_dataframe(u.trajectory.n_frames)
    #         ref_df = pd.read_csv(dict_of_contacts[i].contacts, index_col=[0, 1])
    #         ref_df.columns = ref_df.columns.astype(int)
    #         ref_df = ref_df.astype("int8")
    #         pd.testing.assert_frame_equal(calc_df, ref_df)

    # def test_units_property(self, universes):
    #     for u in universes:
    #         assert u.units in VALID_UNITS

    # def test_units_property_setter(self, universes):
    #     for u in universes:
    #         new_units = "ns"
    #         u.units = new_units
    #         assert u.units == new_units
    #         assert u.units in VALID_UNITS

    # def test_normalize_by_property(self, universes):
    #     for u in universes:
    #         assert u.normalize_by in ["counts", "actual_time", "time_fraction"]

    # def test_normalize_by_property_setter(self, universes):
    #     for u in universes:
    #         new_normalize_by = "counts"
    #         u.normalize_by = new_normalize_by
    #         assert u.normalize_by == new_normalize_by
    #         assert u.normalize_by in ["counts", "actual_time", "time_fraction"]

    # def test_query_property_setter(self, universes):
    #     for u in universes:
    #         new_query = u.select_atoms("protein and (name CA or name BB)")
    #         u.query = new_query
    #         assert u.query is not None
    #         assert isinstance(u.query, ExtendedAtomGroup)

    # def test_database_property_setter(self, universes):
    #     for u in universes:
    #         new_database = u.select_atoms("resname CHOL or resname CHL1")
    #         u.database = new_database
    #         assert u.database is not None
    #         assert isinstance(u.database, ExtendedAtomGroup)



#!/usr/bin/env python

import os
import configparser

# Getting the config file
config = configparser.ConfigParser(allow_no_value=True)
config.read(os.path.join(os.path.abspath(os.path.dirname(__file__)), "../config.ini"))
parameters_config = config["Parameters"]


def get_frame_contact_intervals(frames, tolerance=int(parameters_config["tolerance"])):
    """
    Get frame ranges.

    TODO:
    write doc
    """
    ranges_collect = []
    range_start = 0
    for ix, el in enumerate(frames):
        if ix == 0:
            range_start = el
            continue

        prev_el = frames[ix - 1]
        if not el - tolerance <= prev_el:
            ranges_collect.append((range_start, prev_el))
            range_start = el
        if ix == len(frames) - 1:
            ranges_collect.append((range_start, el))
    return ranges_collect


def calculate_contact_intervals(
    contacts,
    g,
    lipid_id,
    residues_to_show=int(parameters_config["residues_to_show"]),
    intervals_to_filter_out=int(parameters_config["intervals_to_filter_out"]),
):
    """
    TODO:
    write doc
    """
    contact_intervals = {}
    for res, _ in g[lipid_id][:residues_to_show]:
        frame_numbers = contacts.contact_frames[res][lipid_id]
        frame_intervals = get_frame_contact_intervals(frame_numbers)
        for start, end in frame_intervals:
            if end - start < intervals_to_filter_out:
                continue

            if res in contact_intervals:
                contact_intervals[res].append((start, end))
            else:
                contact_intervals[res] = [(start, end)]

    return contact_intervals


def amCharts_contact_intervals(
    contacts,
    g,
    lipid_id,
    residues_to_show=int(parameters_config["residues_to_show"]),
    intervals_to_filter_out=int(parameters_config["intervals_to_filter_out"]),
):
    """
    TODO:
    write doc
    """
    contact_intervals = []
    for res, _ in g[lipid_id][:residues_to_show]:
        frame_numbers = contacts.contact_frames[res][lipid_id]
        frame_intervals = get_frame_contact_intervals(frame_numbers)
        for start, end in frame_intervals:
            if end - start < intervals_to_filter_out:
                continue

            d = {
                "category": res,
                "startFrame": start,
                "endFrame": end,
                "lipid_id": lipid_id,
            }
            contact_intervals.append(d)

    return contact_intervals



from typing import List, Type
import importlib
import inspect

import logging

from prolint2.metrics.base import BaseMetric


class MetricRegistry:
    def __init__(self):
        self._metrics = {}
        self.module_name = "prolint2.metrics.metrics"

        module = importlib.import_module(self.module_name)
        for _, obj in inspect.getmembers(module):
            if (
                inspect.isclass(obj)
                and issubclass(obj, BaseMetric)
                and obj != BaseMetric
            ):
                metric_name = obj.name
                self.register(metric_name, obj)

        # for metric_class in BaseMetric.__subclasses__():
        #     metric_class._register(self)

    def register(self, name: str, metric_class: Type[BaseMetric]):
        if name in self._metrics:
            logging.warning(
                lambda: "Metric with name '%s' already exists in registry.", name
            )
        self._metrics[name] = metric_class

    def get_metric(self, name: str) -> Type[BaseMetric]:
        if name not in self._metrics:
            raise ValueError(f"No metric found with the name '{name}'.")
        return self._metrics[name]

    def get_registered_names(self) -> List[str]:
        return list(self._metrics.keys())



import numpy as np

from prolint2.metrics.base import FittingFunction
from prolint2.metrics.fitters import FittingFunctionFactory


class SurvivalFunction:
    def __init__(self, durations, t_total, delta_t_list):
        self.durations = durations
        self.t_total = t_total
        self.delta_t_list = delta_t_list
        self.num_of_contacts = len(durations)
        self.survival_func = self.calculate()

    def _calc_survival_value(self, delta_t):
        filtered_durations = [
            res_time for res_time in self.durations if res_time >= delta_t
        ]
        sum_res_time = sum(filtered_durations) - delta_t * len(filtered_durations)
        denominator = (self.t_total - delta_t) * self.num_of_contacts

        if delta_t != 0:
            denominator *= self.survival_func0

        return float(sum_res_time) / denominator if denominator != 0 else 0

    def calculate(self):
        survival_func = {}
        for delta_t in self.delta_t_list:
            survival_value = self._calc_survival_value(delta_t)

            if delta_t == 0:
                survival_func[delta_t] = 1
                self.survival_func0 = survival_value
            else:
                survival_func[delta_t] = survival_value

        return survival_func


class KoffCalculator:
    def __init__(
        self, durations, t_total, timestep, fitting_func_name="bi_expo", **kwargs
    ):
        self.durations = durations
        self.t_total = t_total
        self.timestep = timestep
        self.delta_t_list = np.arange(0, t_total, timestep)
        self.kwargs = kwargs

        if self._is_empty_or_zeros(self.durations):
            self.res_time, self.koff = 0, 0
            return

        if fitting_func_name not in FittingFunction.registry:
            func_names = ", ".join(FittingFunction.registry.keys())
            raise ValueError(
                f"Invalid fitting_func_name: {fitting_func_name}. Valid names are: {func_names}"
            )

        self.fitting_func = FittingFunctionFactory.get_fitting_function(
            fitting_func_name
        )
        self.survival_func = SurvivalFunction(
            self.durations, np.max(self.t_total), self.delta_t_list
        ).survival_func
        self.res_time, self.koff = self.calculate_koff()

    def _is_empty_or_zeros(self, array):
        return len(array) == 0 or np.all(array == 0)

    def calculate_koff(self):
        survival_rates = np.nan_to_num(
            [self.survival_func[delta_t] for delta_t in self.delta_t_list]
        )  # TODO: check if nan_to_num is needed
        popt = self.fitting_func.fit(
            np.array(self.delta_t_list), np.array(survival_rates), **self.kwargs
        )
        koff = self.fitting_func.get_koff(popt)
        res_time = 1 / koff
        return res_time, koff



"""
ProLint2: Lipid-Protein Interaction Analysis
============================================

:Authors: Daniel P. Ramirez & Besian I. Sejdiu
:Year: 2022
:Copyright: MIT License

ProLint2 calculates distance-based lipid-protein interactions from molecular dynamics trajectories.
"""

import sys
from setuptools import setup, find_packages
import versioneer

short_description = "ProLint2: Lipid-Protein Interaction Analysis.".split("\n")[0]

# from https://github.com/pytest-dev/pytest-runner#conditional-requirement
needs_pytest = {"pytest", "test", "ptr"}.intersection(sys.argv)
pytest_runner = ["pytest-runner"] if needs_pytest else []

try:
    with open("README.md", "r") as handle:
        long_description = handle.read()
except:
    long_description = None

setup(
    # Self-descriptive entries which should always be present
    name="prolint2",
    author="Daniel P. Ramirez & Besian I. Sejdiu",
    author_email="daniel.ramirezecheme@ucalgary.ca",
    description=short_description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license="MIT",
    # Which Python importable modules should be included when your package is installed
    # Handled automatically by setuptools. Use 'exclude' to prevent some specific
    # subpackage(s) from being added, if needed
    packages=find_packages(),
    # Optional include package data to ship with your package
    # Customize MANIFEST.in if the general case does not suit your needs
    # Comment out this line to prevent the files from being packaged with your software
    include_package_data=True,
    # Allows `setup.py test` to work correctly with pytest
    setup_requires=[] + pytest_runner,
    scripts=[
        "bin/prolint2",
    ],
    # Additional entries you may want simply uncomment the lines you want and fill in the data
    url="https://github.com/ProLint/prolint2",  # Website
    install_requires=[
        "pandas",
        "mdanalysis",
        "bottle",
        "typer[all]",
        "requests",
        "networkx",
        "logomaker",
        "seaborn",
        "typing_extensions",
    ],  # Required packages, pulls from pip if needed; do not use for Conda deployment
    platforms=[
        "Linux",
        "Mac OS-X",
        "Windows",
    ],  # Valid platforms your code works on, adjust to your flavor
    python_requires=">=3.10",  # Python version restrictions
    # Manual control if final package is compressible or not, set False to prevent the .egg from being made
    # zip_safe=False,
)
